//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <iostream>
#include <fstream>
// #include <random>
#include <yaml-cpp/yaml.h>
#include <Eigen/Dense>
#include <ctime>
#include "eigenmvn.h"

#include "pinocchio/parsers/urdf.hpp"
#include "pinocchio/algorithm/joint-configuration.hpp"
#include "pinocchio/algorithm/kinematics.hpp"
#include "pinocchio/algorithm/center-of-mass.hpp"
#include "pinocchio/algorithm/centroidal.hpp"
#include "pinocchio/algorithm/rnea.hpp"
#include "pinocchio/algorithm/crba.hpp"
#include "pinocchio/algorithm/frames.hpp"

// CNPY
#include<cnpy.h>

// WOLF
#include <core/problem/problem.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/capture/capture_base.h>
#include <core/capture/capture_pose.h>
#include <core/feature/feature_base.h>
#include <core/factor/factor_block_absolute.h>

// IMU
#include "imu/sensor/sensor_imu.h"
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu.h"

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/sensor/sensor_inertial_kinematics.h"
#include "bodydynamics/sensor/sensor_force_torque.h"
#include "bodydynamics/capture/capture_inertial_kinematics.h"
#include "bodydynamics/capture/capture_force_torque_.h"
#include "bodydynamics/capture/capture_leg_odom.h"
#include "bodydynamics/processor/processor_inertial_kinematics.h"
#include "bodydynamics/processor/processor_force_torque_.h"
#include "bodydynamics/factor/factor_inertial_kinematics.h"
#include "bodydynamics/factor/factor_force_torque_.h"

#include "bodydynamics/capture/capture_point_feet_nomove.h"
#include "bodydynamics/sensor/sensor_point_feet_nomove.h"
#include "bodydynamics/processor/processor_point_feet_nomove.h"

#include "mcapi_utils.h"


using namespace Eigen;
using namespace wolf;
using namespace pinocchio;

typedef pinocchio::SE3Tpl<double> SE3;
typedef pinocchio::ForceTpl<double> Force;


int main (int argc, char **argv) {
    // retrieve parameters from yaml file
    YAML::Node config = YAML::LoadFile("/home/mfourmy/Documents/Phd_LAAS/wolf/bodydynamics/demos/solo_real_estimation.yaml");
    std::string robot_urdf = config["robot_urdf"].as<std::string>();
    int dimc = config["dimc"].as<int>();
    double mass_offset = config["mass_offset"].as<double>();
    double dt = config["dt"].as<double>();
    // double min_t = config["min_t"].as<double>();
    double max_t = config["max_t"].as<double>();
    double solve_every_sec = config["solve_every_sec"].as<double>();
    bool solve_end = config["solve_end"].as<bool>();

    // Ceres setup
    int max_num_iterations = config["max_num_iterations"].as<int>();

    // estimator sensor noises
    double std_pbc_est    = config["std_pbc_est"].as<double>();
    double std_vbc_est    = config["std_vbc_est"].as<double>();
    double std_f_est      = config["std_f_est"].as<double>();
    double std_tau_est    = config["std_tau_est"].as<double>();
    double std_foot_nomove = config["std_foot_nomove"].as<double>();
    
    // priors
    double std_prior_p = config["std_prior_p"].as<double>();
    double std_prior_o = config["std_prior_o"].as<double>();
    double std_prior_v = config["std_prior_v"].as<double>();
    double std_abs_bias_pbc  = config["std_abs_bias_pbc"].as<double>();
    double std_abs_bias_acc  = config["std_abs_bias_acc"].as<double>();
    double std_abs_bias_gyro = config["std_abs_bias_gyro"].as<double>();    
    double std_bp_drift      = config["std_bp_drift"].as<double>();
    std::vector<double> bias_pbc_prior_vec = config["bias_pbc_prior"].as<std::vector<double>>();
    Eigen::Map<Vector3d> bias_pbc_prior(bias_pbc_prior_vec.data());
    std::vector<double> bias_imu_prior_vec = config["bias_imu_prior"].as<std::vector<double>>();
    Eigen::Map<Vector6d> bias_imu_prior(bias_imu_prior_vec.data());

    double fz_thresh = config["fz_thresh"].as<double>();
    std::vector<double> b_p_bi_vec = config["b_p_bi"].as<std::vector<double>>();
    std::vector<double> b_q_i_vec = config["b_q_i"].as<std::vector<double>>();
    Eigen::Map<Vector3d> b_p_bi(b_p_bi_vec.data());
    Eigen::Map<Vector4d> b_qvec_i(b_q_i_vec.data());

    std::vector<double> artificial_bias_p_vec = config["artificial_bias_p"].as<std::vector<double>>();
    Eigen::Map<Vector3d> artificial_bias_p(artificial_bias_p_vec.data());

    std::string data_file_path = config["data_file_path"].as<std::string>();
    std::string out_npz_file_path = config["out_npz_file_path"].as<std::string>();

    std::vector<std::string> ee_names = {"FL_ANKLE", "FR_ANKLE", "HL_ANKLE", "HR_ANKLE"};
    unsigned int nbc = ee_names.size();
    // Base to IMU transform
    Quaterniond b_q_i(b_qvec_i);
    SE3 b_T_i(b_q_i, b_p_bi);
    SE3 i_T_b = b_T_i.inverse();
    Matrix3d b_R_i =  b_T_i.rotation();
    Vector3d i_p_ib = i_T_b.translation();
    Matrix3d i_R_b =  i_T_b.rotation();

    // initialize some vectors and fill them with dicretized data
    cnpy::npz_t arr_map = cnpy::npz_load(data_file_path);

    //load it into a new array
    cnpy::NpyArray t_npyarr = arr_map["t"];
    cnpy::NpyArray imu_acc_npyarr = arr_map["imu_acc"];
    // cnpy::NpyArray o_a_oi_npyarr = arr_map["o_a_oi"];
    cnpy::NpyArray o_q_i_npyarr = arr_map["o_q_i"];
    cnpy::NpyArray i_omg_oi_npyarr = arr_map["i_omg_oi"];
    cnpy::NpyArray qa_npyarr = arr_map["qa"];
    cnpy::NpyArray dqa_npyarr = arr_map["dqa"];
    cnpy::NpyArray tau_npyarr = arr_map["tau"];
    cnpy::NpyArray l_forces_npyarr = arr_map["l_forces"];
    // cnpy::NpyArray w_pose_wm_npyarr = arr_map["w_pose_wm"];
    cnpy::NpyArray m_v_wm_npyarr = arr_map["m_v_wm"];
    cnpy::NpyArray w_v_wm_npyarr = arr_map["w_v_wm"];
    // cnpy::NpyArray o_R_i_npyarr = arr_map["o_R_i"];
    cnpy::NpyArray w_p_wm_npyarr = arr_map["w_p_wm"];
    cnpy::NpyArray w_q_m_npyarr = arr_map["w_q_m"];
    // cnpy::NpyArray w_R_m_npyarr = arr_map["w_R_m"];
    double* t_arr = t_npyarr.data<double>();
    double* imu_acc_arr = imu_acc_npyarr.data<double>();
    // double* o_a_oi_arr = o_a_oi_npyarr.data<double>();
    double* i_omg_oi_arr = i_omg_oi_npyarr.data<double>();
    double* qa_arr = qa_npyarr.data<double>();
    double* dqa_arr = dqa_npyarr.data<double>();
    double* tau_arr = tau_npyarr.data<double>();
    double* l_forces_arr = l_forces_npyarr.data<double>();
    // double* w_pose_wm_arr = w_pose_wm_npyarr.data<double>();
    double* m_v_wm_arr = m_v_wm_npyarr.data<double>();
    double* w_v_wm_arr = w_v_wm_npyarr.data<double>();
    double* o_q_i_arr = o_q_i_npyarr.data<double>();
    // double* o_R_i_arr = o_R_i_npyarr.data<double>();
    double* w_p_wm_arr = w_p_wm_npyarr.data<double>();
    double* w_q_m_arr = w_q_m_npyarr.data<double>();
    // double* w_R_m_arr = w_R_m_npyarr.data<double>();
    unsigned int N = t_npyarr.shape[0];
    if (max_t > 0){
        N = N < int(max_t/dt) ? N : int(max_t/dt);
    }

    // Creating measurements from simulated trajectory
    // Load the urdf model
    Model model;
    pinocchio::urdf::buildModel(robot_urdf, JointModelFreeFlyer(), model);
    std::cout << "model name: " << model.name << std::endl;
    // set gravity of the model to the one decided by wolf
    model.gravity.linear() = wolf::gravity();
    // to accomodate the difference between real robot and model masses
    model.inertias[1].mass() = model.inertias[1].mass() + mass_offset;

    // add artificial lever to test bias estimation on real data
    VectorXd q_toto(19); q_toto << 0,0,0, 0,0,0,1, 0,0,0,0,0,0,0,0,0,0,0,0; // TODO: put current base orientation estimate? YES

    Data data(model);
    
    centerOfMass(model, data, q_toto);        
    std::cout << "artificial_bias_p: " << artificial_bias_p.transpose() << std::endl;
    std::cout << "model.inertias[1].lever(): " << model.inertias[1].lever().transpose() << std::endl;
    std::cout << "data.com[0]: " << data.com[0].transpose() << std::endl;
    model.inertias[1] = pinocchio::Inertia(model.inertias[1].mass(), artificial_bias_p, model.inertias[1].inertia());
    centerOfMass(model, data, q_toto);        
    std::cout << "model.inertias[1].lever(): " << model.inertias[1].lever().transpose() << std::endl;
    std::cout << "data.com[0]: " << data.com[0].transpose() << std::endl;


    // Recover end effector frame ids
    std::vector<int> ee_ids;
    std::cout << "Frame ids" << std::endl;
    for (auto ee_name: ee_names){
        ee_ids.push_back(model.getFrameId(ee_name));
        std::cout << ee_name << std::endl;
    }

    /////////////////////////
    // WOLF enters the scene
    // SETUP PROBLEM/SENSORS/PROCESSORS
    std::string bodydynamics_root_dir = _WOLF_BODYDYNAMICS_ROOT_DIR;

    ProblemPtr problem = Problem::create("POVCDL", 3);

    //////////////////////
    // COMPUTE INITIAL STATE
    TimeStamp t0(t_arr[0]);
    Eigen::Map<Matrix<double,12, 1>> qa(qa_arr);
    Eigen::Map<Matrix<double,12, 1>> dqa(dqa_arr);
    Eigen::Map<Vector4d> o_q_i(o_q_i_arr);  // initialize with IMU orientation
    Eigen::Map<Vector3d> i_omg_oi(i_omg_oi_arr);  // Hyp: b=i (not the case)
    Eigen::Map<Vector3d> w_p_wm(w_p_wm_arr);
    Eigen::Map<Vector4d> w_qvec_wm(w_q_m_arr);
    w_qvec_wm.normalize();  // not close enough to wolf eps sometimes

    // initial state
    VectorXd q(19); q << 0,0,0, o_q_i, qa; // TODO: put current base orientation estimate? YES
    VectorXd dq(18); dq << 0,0,0, i_omg_oi, dqa; // TODO: put current base velocity estimate? YES


    centerOfMass(model, data, q, dq);        
    // Vector3d w_p_wc = data.com[0]; 
    // Vector3d w_v_wc = data.vcom[0];

    // gesticulation in base frame: just compute centroidal momentum with static q and vq
    // Vector3d w_Lc = computeCentroidalMomentum(model, data, q, dq).angular();

    // hyp: robot not moving initially (no vel and angular momentum)
    VectorXd x_origin; x_origin.resize(19);
    // x_origin << w_p_wm, w_qvec_wm, 0,0,0, w_p_wc, 0,0,0, 0,0,0;
    x_origin << 0,0,0, o_q_i, 0,0,0, 0,0,0, 0,0,0, 0,0,0;
    //////////////////////
    //////////////////////

    // CERES WRAPPER
    SolverCeresPtr solver = std::make_shared<SolverCeres>(problem);
    solver->getSolverOptions().max_num_iterations = max_num_iterations;
    
    //===================================================== INITIALIZATION
    // SENSOR + PROCESSOR INERTIAL KINEMATICS
    ParamsSensorInertialKinematicsPtr intr_ik = std::make_shared<ParamsSensorInertialKinematics>();
    intr_ik->std_pbc = std_pbc_est;
    intr_ik->std_vbc = std_vbc_est;
    VectorXd extr0; // default size for dynamic vector is 0-0 -> what we need
    SensorBasePtr sen_ikin_base = problem->installSensor("SensorInertialKinematics", "SenIK", extr0, intr_ik);
    // SensorBasePtr sen_ikin_base = problem->installSensor("SensorInertialKinematics", "SenIK", extr,  bodydynamics_root_dir + "/demos/sensor_inertial_kinematics.yaml");  // TODO: does not work!
    SensorInertialKinematicsPtr sen_ikin = std::static_pointer_cast<SensorInertialKinematics>(sen_ikin_base);

    ParamsProcessorInertialKinematicsPtr params_ik = std::make_shared<ParamsProcessorInertialKinematics>();
    params_ik->sensor_angvel_name = "SenImu";
    params_ik->std_bp_drift = std_bp_drift;
    ProcessorBasePtr proc_ikin_base = problem->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", sen_ikin, params_ik);
    // ProcessorBasePtr proc_ikin_base = problem->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", "SenIK", bodydynamics_root_dir + "/demos/processor_inertial_kinematics.yaml");
    ProcessorInertialKinematicsPtr proc_inkin = std::static_pointer_cast<ProcessorInertialKinematics>(proc_ikin_base);

    // SENSOR + PROCESSOR IMU
    SensorBasePtr sen_imu_base = problem->installSensor("SensorImu", "SenImu", (Vector7d() << 0, 0, 0, 0, 0, 0, 1).finished(), bodydynamics_root_dir + "/demos/sensor_imu_solo12.yaml");
    SensorImuPtr sen_imu = std::static_pointer_cast<SensorImu>(sen_imu_base);
    ProcessorBasePtr proc_ftp_base = problem->installProcessor("ProcessorImu", "imu pre-integrator", "SenImu", bodydynamics_root_dir + "/demos/processor_imu_solo12.yaml");
    ProcessorImuPtr proc_imu = std::static_pointer_cast<ProcessorImu>(proc_ftp_base);

    // SENSOR + PROCESSOR FT PREINT
    ParamsSensorForceTorquePtr intr_ft = std::make_shared<ParamsSensorForceTorque>();
    intr_ft->std_f = std_f_est;
    intr_ft->std_tau = std_tau_est;
    intr_ft->mass = data.mass[0];
    std::cout << std::setprecision(std::numeric_limits<long double>::digits10 + 1) << "\n\nROBOT MASS: " << intr_ft->mass << std::endl;
    SensorBasePtr sen_ft_base = problem->installSensor("SensorForceTorque", "SenFT", VectorXd(0), intr_ft);
    // SensorBasePtr sen_ft_base = problem->installSensor("SensorForceTorque", "SenFT", VectorXd(0), bodydynamics_root_dir + "/demos/sensor_ft.yaml");
    SensorForceTorquePtr sen_ft = std::static_pointer_cast<SensorForceTorque>(sen_ft_base);
    ParamsProcessorForceTorquePtr params_sen_ft = std::make_shared<ParamsProcessorForceTorque>();
    params_sen_ft->sensor_ikin_name = "SenIK";
    params_sen_ft->sensor_angvel_name = "SenImu";
    params_sen_ft->nbc = nbc;
    params_sen_ft->dimc = dimc;
    params_sen_ft->time_tolerance = 0.0005;
    params_sen_ft->unmeasured_perturbation_std = 0.000001;
    params_sen_ft->max_time_span = 1000;
    params_sen_ft->max_buff_length = 500;
    params_sen_ft->dist_traveled = 20000.0;
    params_sen_ft->angle_turned = 1000;
    params_sen_ft->voting_active = false;
    ProcessorBasePtr proc_ft_base = problem->installProcessor("ProcessorForceTorque", "PFT", sen_ft, params_sen_ft);
    // ProcessorBasePtr proc_ft_base = problem->installProcessor("ProcessorForceTorque", "PFT", "SenFT", bodydynamics_root_dir + "/demos/processor_ft_preint.yaml");
    ProcessorForceTorquePtr proc_ft = std::static_pointer_cast<ProcessorForceTorque>(proc_ft_base);

    // SENSOR + PROCESSOR POINT FEET NOMOVE
    ParamsSensorPointFeetNomovePtr intr_pfn = std::make_shared<ParamsSensorPointFeetNomove>();
    intr_pfn->std_foot_nomove_ = std_foot_nomove;
    VectorXd extr;  // default size for dynamic vector is 0-0 -> what we need
    SensorBasePtr sen_pfnm_base = problem->installSensor("SensorPointFeetNomove", "SenPfnm", extr, intr_pfn);
    SensorPointFeetNomovePtr sen_pfnm = std::static_pointer_cast<SensorPointFeetNomove>(sen_pfnm_base);
    ParamsProcessorPointFeetNomovePtr params_pfnm = std::make_shared<ParamsProcessorPointFeetNomove>();
    ProcessorBasePtr proc_pfnm_base = problem->installProcessor("ProcessorPointFeetNomove", "PPointFeetNomove", sen_pfnm, params_pfnm);


    // SETPRIOR RETRO-ENGINEERING
    // We are not using setPrior because of processors initial captures problems so we have to
    // - Manually create the first KF and its pose and velocity priors
    // - call setOrigin on processors MotionProvider since the flag prior_is_set is not true when doing installProcessor (later)

    MatrixXd P_origin(9,9); P_origin.setIdentity();
    P_origin.block<3,3>(0,0)   = pow(std_prior_p, 2) * Matrix3d::Identity();
    P_origin.block<3,3>(3,3)   = pow(std_prior_o, 2) * Matrix3d::Identity();
    P_origin.block<3,3>(6,6)   = pow(std_prior_v, 2) * Matrix3d::Identity();
    FrameBasePtr KF1 = problem->emplaceFrame(t0, x_origin);
    // Prior pose factor
    CapturePosePtr pose_prior_capture = CaptureBase::emplace<CapturePose>(KF1, t0, nullptr, x_origin.head(7), P_origin.topLeftCorner(6, 6));
    pose_prior_capture->emplaceFeatureAndFactor();
    // Prior velocity factor
    CaptureBasePtr capV0 = CaptureBase::emplace<CaptureBase>(KF1, "Vel0", t0);
    FeatureBasePtr featV0 = FeatureBase::emplace<FeatureBase>(capV0, "Vel0", x_origin.segment<3>(7), P_origin.block<3, 3>(6, 6));
    FactorBasePtr facV0 = FactorBase::emplace<FactorBlockAbsolute>(featV0, featV0, KF1->getV(), nullptr, false);

    // Special trick to make things work in the IMU + IKin + FTPreint case
    // 0. Call keyFrameCallback of processorIKin so that its kf_buffer contains the first KF0
    // 1. Call setOrigin processorIMU  to generate a fake captureIMU -> generates b_imu
    // 2. process one IKin capture corresponding to the same KF, the factor needing b_imu -> generates a b_pbc
    // 3. set processorForceTorquePreint origin which requires both b_imu and b_pbc
    proc_inkin->keyFrameCallback(KF1);
    proc_imu->setOrigin(KF1);
    // proc_legodom->setOrigin(KF1);

    // INERTIAL KINEMATICS CAPTURE DATA ZERO MOTION
    VectorXd q_static = q;    q_static.head<7>() << 0,0,0, 0,0,0,1;
    VectorXd dq_static = dq; dq_static.head<6>() << 0,0,0, 0,0,0;
    centerOfMass(model, data, q_static, dq_static);        
    Vector3d b_p_bc = data.com[0];  // meas   
    Vector3d i_p_ic = i_T_b.act(b_p_bc);
    Vector3d i_v_ic = i_R_b*data.vcom[0];  // meas
    Matrix3d b_Iw = compute_Iw(model, data, q_static, b_p_bc);
    Matrix3d i_Iw = i_R_b*b_Iw*b_R_i;  // !! rotate on both sides to keep a symmetric matrix     
    // Vector3d b_Lc_gesti = computeCentroidalMomentum(model, data, q_static, dq_static).angular();
    auto b_centro_momentum_gesti = computeCentroidalMomentum(model, data, q_static, dq_static);
    auto i_centro_momentum_gesti = i_T_b.act(b_centro_momentum_gesti);
    Vector3d i_Lc_gesti = i_centro_momentum_gesti.angular();

    // momentum parameters
    VectorXd meas_ikin(9); meas_ikin << i_p_ic, i_v_ic, i_omg_oi;
    auto CIKin0 = std::make_shared<CaptureInertialKinematics>(t0, sen_ikin, meas_ikin, i_Iw, i_Lc_gesti);
    CIKin0->process();
    proc_ft->setOrigin(KF1);

    ////////////////////////////////////////////
    // INITIAL BIAS FACTORS
    // Add absolute factor on Imu biases to simulate a fix()
    Vector6d std_abs_imu; std_abs_imu << std_abs_bias_acc*Vector3d::Ones(), std_abs_bias_gyro*Vector3d::Ones();
    Matrix6d Q_bi = std_abs_imu.array().square().matrix().asDiagonal();    
    CaptureBasePtr capbi0 = CaptureBase::emplace<CaptureBase>(KF1, "bi0", t0);
    FeatureBasePtr featbi0 = FeatureBase::emplace<FeatureBase>(capbi0, "bi0", bias_imu_prior, Q_bi);
    FactorBasePtr facbi0 = FactorBase::emplace<FactorBlockAbsolute>(featbi0, featbi0, KF1->getCaptureOf(sen_imu)->getSensorIntrinsic(), nullptr, false);  
    sen_imu->getStateBlock('I')->setState(bias_imu_prior);

    // Add loose absolute factor on initial bp bias since dynamical trajectories 
    Matrix3d Q_bp = pow(std_abs_bias_pbc, 2)* Matrix3d::Identity();
    CaptureBasePtr cap_bp0 = CaptureBase::emplace<CaptureBase>(KF1, "bp0", t0);
    FeatureBasePtr feat_bp0 = FeatureBase::emplace<FeatureBase>(cap_bp0, "bp0", bias_pbc_prior, Q_bp);
    FactorBasePtr fac_bp0 = FactorBase::emplace<FactorBlockAbsolute>(feat_bp0, feat_bp0, KF1->getCaptureOf(sen_ikin)->getSensorIntrinsic(), nullptr, false); 
    sen_ikin->getStateBlock('I')->setState(bias_pbc_prior);


    ///////////////////////////////////////////
    // process measurements in sequential order
    double ts_since_last_kf = 0;

    // for point feet factor
    std::vector<Vector3d> i_p_il_vec_origin;
    FrameBasePtr KF_origin = KF1;

    //////////////////////////////////////////
    // Vectors to store estimates at the time of data processing 
    // fbk -> feedback: to check if the estimation is stable enough for feedback control
    std::vector<Vector3d> p_ob_fbk_v; 
    std::vector<Vector4d> q_ob_fbk_v; 
    std::vector<Vector3d> v_ob_fbk_v; 
    std::vector<Vector3d> c_ob_fbk_v; 
    std::vector<Vector3d> cd_ob_fbk_v; 
    std::vector<Vector3d> Lc_ob_fbk_v; 

    std::vector<Vector3d> i_omg_oi_v; 
    //////////////////////////////////////////

    unsigned int nb_kf = 1;
    for (unsigned int i=0; i < N; i++){
        TimeStamp ts(t_arr[i]);

        ////////////////////////////////////
        // Retrieve current state 
        VectorComposite curr_state = problem->getState();
        Vector4d o_qvec_i_curr = curr_state['O'];
        Quaterniond o_q_i_curr(o_qvec_i_curr); 
        Vector3d o_v_oi_curr = curr_state['V']; 

        ////////////////////////////////////
        ////////////////////////////////////
        // Get measurements
        // retrieve traj data
        Eigen::Map<Vector3d> imu_acc(imu_acc_arr+3*i);
        Eigen::Map<Vector3d> i_omg_oi(i_omg_oi_arr+3*i);  // Hyp: b=i (not the case)
        Eigen::Map<Matrix<double,12, 1>> tau(tau_arr+12*i);
        Eigen::Map<Matrix<double,12, 1>> qa(qa_arr+12*i);
        Eigen::Map<Matrix<double,12, 1>> dqa(dqa_arr+12*i);

        // store i_omg_oi for later computation of o_v_ob from o_v_oi
        i_omg_oi_v.push_back(i_omg_oi);

        ////////////////////////////////////
        // IMU
        // imu_acc << -wolf::gravity(); 
        // i_omg_oi << 0,0,0;
        Vector6d acc_gyr_meas; acc_gyr_meas << imu_acc, i_omg_oi;
        Matrix6d acc_gyr_cov = sen_imu->getNoiseCov();

        ////////////////////////////////////
        // Kinematics + forces
        SE3 o_T_i(o_q_i_curr, Vector3d::Zero()); // TODO: put current orientation estimate here
        Matrix3d o_R_i = o_T_i.rotation();
        Matrix3d i_R_o = o_R_i.transpose();
        Vector3d i_v_oi_curr = i_R_o * o_v_oi_curr; 
        Vector3d i_acc = imu_acc + i_R_o * gravity();
        Vector3d i_asp_i = i_acc - i_omg_oi.cross(i_v_oi_curr);

        VectorXd q(19); q << 0,0,0, o_q_i_curr.coeffs(), qa;
        // VectorXd dq(18); dq << i_v_oi_curr, i_omg_oi, dqa;
        VectorXd dq(18); dq << 0,0,0, i_omg_oi, dqa;
        VectorXd ddq(model.nv); ddq.setZero();
        ddq.head<3>() = i_asp_i;
        // TODO: add other terms to compute forces (ddqa...)

        // update and retrieve kinematics
        forwardKinematics(model, data, q);
        updateFramePlacements(model, data);

        // compute all linear jacobians in feet frames (only for quadruped)
        MatrixXd Jlinvel(12, model.nv); Jlinvel.setZero();
        for (unsigned int i_ee=0; i_ee < nbc; i_ee++){
            MatrixXd Jee(6, model.nv); Jee.setZero();
            computeFrameJacobian(model, data, q, ee_ids[i_ee], Jee);
            Jlinvel.block(3*i_ee, 0, 3, model.nv) = Jee.topRows<3>();
        }

        // estimate forces from torques
        VectorXd tau_cf = rnea(model, data, q, dq, ddq);  // compute total torques applied on the joints (and free flyer)
        tau_cf.tail(model.nv-6) -= tau;  // remove measured joint torque (not on the free flyer)

        // VectorXd forces = Jlinvel_reduced.transpose().lu().solve(tau_joints); // works only for exactly determined system, removing free flyer from the system -> 12x12 J mat 
        // Solve in Least square sense (3 ways):
        // VectorXd forces = Jlinvel.transpose().bdcSvd(Eigen::ComputeThinU | Eigen::ComputeThinV).solve(tau_cf);  // SVD
        // VectorXd l_forces = Jlinvel.transpose().colPivHouseholderQr().solve(tau_cf);  // QR
        // (A.transpose() * A).ldlt().solve(A.transpose() * b)  // solve the normal equation (twice less accurate than other 2)

        // Or else, retrieve from preprocessed dataset
        Eigen::Map<Matrix<double,12, 1>> l_forces(l_forces_arr+12*i);


        Vector3d o_f_tot = Vector3d::Zero();
        std::vector<Vector3d> l_f_l_vec;
        std::vector<Vector3d> o_f_l_vec;
        std::vector<Vector3d> i_p_il_vec;
        std::vector<Vector4d> i_qvec_l_vec;
        // needing relative kinematics measurements
        VectorXd q_static = q; q_static.head<7>() << 0,0,0, 0,0,0,1;
        forwardKinematics(model, data, q_static);
        updateFramePlacements(model, data);
        for (unsigned int i_ee=0; i_ee < nbc; i_ee++){
            auto b_T_l = data.oMf[ee_ids[i_ee]];
            auto i_T_l = i_T_b * b_T_l;
            Vector3d i_p_il = i_T_l.translation();                       // meas
            Vector4d i_qvec_l = Quaterniond(i_T_l.rotation()).coeffs();  // meas
            Vector3d l_f_l = l_forces.segment(3*i_ee, 3);                // meas
            Vector3d o_f_l = o_R_i*i_T_l.rotation() * l_f_l;     // for contact test
            o_f_tot += o_f_l;

            // // put the measure to a constant for unit testing 
            // i_p_il << 0,0,0;
            // i_qvec_l << 0,0,0,1;
            // l_f_l << -(intr_ft->mass-0.01)*wolf::gravity()/4;

            l_f_l_vec.push_back(l_f_l);
            o_f_l_vec.push_back(o_f_l);
            i_p_il_vec.push_back(i_p_il);
            i_qvec_l_vec.push_back(i_qvec_l);
        }
        // i_p_il_vec[0] <<  1, 0, 0;
        // i_p_il_vec[1] << -1, 0, 0;
        // i_p_il_vec[2] <<  0, 1, 0;
        // i_p_il_vec[3] <<  0,-1, 0;

        // initialization for point feet factor
        if (i == 0){
            i_p_il_vec_origin = i_p_il_vec;
        }
        // std::cout << "o_f_tot:       " << o_f_tot.transpose() << std::endl;
        // std::cout << "o_f_tot + m*g: " << (o_f_tot + data.mass[0]*model.gravity.linear()).transpose() << std::endl;


        //////////////////////////////////////////////
        // COM relative position and velocity measures
        centerOfMass(model, data, q_static, dq_static);        
        Vector3d b_p_bc = data.com[0];  
        // std::cout << "b_p_bc: " << b_p_bc.transpose() << std::endl; 
        Vector3d i_p_ic = i_T_b.act(b_p_bc);  // meas
        // velocity due to gesticulation only in base frame 
        // -> set world frame = base frame and set body frame spatial vel to zero 
        Vector3d i_v_ic = i_R_b*data.vcom[0];  // meas
        // Centroidal inertia and gesticulation kinetic momentum 
        Matrix3d b_Iw = compute_Iw(model, data, q_static, b_p_bc);
        Matrix3d i_Iw = i_R_b*b_Iw*b_R_i;  // !! rotate on both sides to keep a symmetric matrix 
        // gesticulation in base frame: just compute centroidal momentum with static q and vq
        // Vector3d b_Lc_gesti = computeCentroidalMomentum(model, data, q_static, dq_static).angular();
        auto b_centro_momentum_gesti = computeCentroidalMomentum(model, data, q_static, dq_static);
        auto i_centro_momentum_gesti = i_T_b.act(b_centro_momentum_gesti);
        Vector3d i_Lc_gesti = i_centro_momentum_gesti.angular();

        // Inertial kinematics
        meas_ikin << i_p_ic, i_v_ic, i_omg_oi;

        ////////////////////
        // Force Torque
        ////////////////////

        VectorXd f_meas(nbc*3); 
        VectorXd tau_meas(nbc*3);
        VectorXd kin_meas(nbc*7);
        for (unsigned int i_ee=0; i_ee < nbc; i_ee++){
            f_meas.segment<3>(3*i_ee) = l_f_l_vec[i_ee]; 
            kin_meas.segment<3>(i_ee*3)         = i_p_il_vec[i_ee];
            kin_meas.segment<4>(nbc*3 + i_ee*4) = i_qvec_l_vec[i_ee];
        }

        VectorXd cap_ftp_data(dimc*nbc+3+3+nbc*7); 
        MatrixXd Qftp(dimc*nbc+3+3,dimc*nbc+3+3);  // kin not in covariance mat
        cap_ftp_data << f_meas, i_p_ic, i_omg_oi, kin_meas;
        Qftp.setIdentity();
        Qftp.block(0,0, nbc*3,nbc*3) *= pow(std_f_est, 2);
        Qftp.block<3, 3>(nbc*dimc, nbc*dimc, 3,3) *= pow(sen_ikin->getPbcNoiseStd(), 2);
        Qftp.block<3, 3>(nbc*dimc+3,nbc*dimc+3, 3,3) = acc_gyr_cov.bottomRightCorner<3,3>();
        ////////////////////

        CaptureImuPtr CImu = std::make_shared<CaptureImu>(ts, sen_imu, acc_gyr_meas, acc_gyr_cov);
        CImu->process();
        auto CIKin = std::make_shared<CaptureInertialKinematics>(ts, sen_ikin, meas_ikin, i_Iw, i_Lc_gesti);
        CIKin->process();
        auto CFTpreint = std::make_shared<CaptureForceTorque>(ts, sen_ft, CIKin, CImu, cap_ftp_data, Qftp);
        CFTpreint->process();


        // 1) detect feet in contact
        std::vector<int> feet_in_contact;
        int nb_feet_in_contact = 0;
        std::unordered_map<int, Vector3d> kin_incontact;
        for (unsigned int i_ee=0; i_ee<nbc; i_ee++){
            // basic contact detection based on normal foot vel, things break if no foot is in contact
            // std::cout << "F" << ee_names[i_ee] << " norm: " << wrench_meas_v[i_ee][i].norm() << std::endl;
            // std::cout << "o_f_l_vec[i_ee]: " << o_f_l_vec[i_ee].transpose() << std::endl;
            if (o_f_l_vec[i_ee](2) > fz_thresh){
                feet_in_contact.push_back(i_ee);
                nb_feet_in_contact++;
                kin_incontact.insert({i_ee, i_p_il_vec[i_ee]});
            }
        }

        // TODO: ???? kin_incontact is a map to Vector3d in this implementation, should be Vector7d [p, q]
        // CapturePointFeetNomovePtr CPF = std::make_shared<CapturePointFeetNomove>(ts, sen_pfnm, kin_incontact);
        // CPF->process();




        // add zero vel artificial factor
        if (problem->getTrajectory()->size() > nb_kf)
        {
            auto kf = problem->getTrajectory()->getLastFrame();
            std::cout << "New KF " << kf->getTimeStamp() << std::endl;

            // CaptureBasePtr capV0 = CaptureBase::emplace<CaptureBase>(kf, "Vel0", kf->getTimeStamp());
            // FeatureBasePtr featV0 = FeatureBase::emplace<FeatureBase>(capV0, "Vel0", Vector3d::Zero(), 0.00001*Matrix3d::Ones());
            // FactorBasePtr facV0 = FactorBase::emplace<FactorBlockAbsolute>(featV0, featV0, kf->getV(), nullptr, false);

            nb_kf++;
            KF_origin = kf;
        }



        ts_since_last_kf += dt;
        if ((solve_every_sec > 0) && (ts_since_last_kf >= solve_every_sec)){
            // problem->print(4,true,true,true);
            std::string report = solver->solve(SolverManager::ReportVerbosity::BRIEF);
            std::cout << report << std::endl;
            ts_since_last_kf = 0;
        }

        // Store current state estimation
        VectorComposite state_fbk = problem->getState(ts);
        Vector3d o_p_oi_new = state_fbk['P'];
        Vector4d o_qvec_i_new = state_fbk['O'];
        Vector3d o_v_oi_new = state_fbk['V'];

        Matrix3d o_R_i_new = Quaterniond(o_qvec_i_new).toRotationMatrix();
        Matrix3d o_R_b_new = o_R_i_new * i_R_b;
        Vector4d o_qvec_b_new = Quaterniond(o_R_b_new).coeffs();
        Vector3d o_p_ob_new = o_p_oi_new + o_R_i_new * i_p_ib;
        Vector3d o_v_ob_new = o_v_oi_new + o_R_b_new * i_omg_oi.cross(b_p_bi);

        p_ob_fbk_v.push_back(o_p_ob_new);
        q_ob_fbk_v.push_back(o_qvec_b_new);
        v_ob_fbk_v.push_back(o_v_ob_new);
        c_ob_fbk_v.push_back(state_fbk['C']);
        cd_ob_fbk_v.push_back(state_fbk['D']);
        Lc_ob_fbk_v.push_back(state_fbk['L']);
    }
    


    ///////////////////////////////////////////
    //////////////// SOLVING //////////////////
    ///////////////////////////////////////////
    problem->print(4,true,true,true);
    if (solve_end){
        std::string report = solver->solve(SolverManager::ReportVerbosity::BRIEF);
        problem->print(4,true,true,true);
        std::cout << report << std::endl;
    }

    //////////////////////////////////////
    //////////////////////////////////////
    //            STORE DATA            //
    //////////////////////////////////////
    //////////////////////////////////////
    std::fstream file_est; 
    std::fstream file_fbk; 
    std::fstream file_kfs; 
    std::fstream file_cov; 
    file_est.open("/home/mfourmy/Documents/Phd_LAAS/phd_misc/centroidalkin/est.csv", std::fstream::out); 
    file_fbk.open("/home/mfourmy/Documents/Phd_LAAS/phd_misc/centroidalkin/fbk.csv", std::fstream::out); 
    file_kfs.open("/home/mfourmy/Documents/Phd_LAAS/phd_misc/centroidalkin/kfs.csv", std::fstream::out); 
    file_cov.open("/home/mfourmy/Documents/Phd_LAAS/phd_misc/centroidalkin/cov.csv", std::fstream::out); 
    std::string header_est = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,cx,cy,cz,cdx,cdy,cdz,Lx,Ly,Lz\n";
    file_est << header_est;
    file_fbk << header_est;
    std::string header_kfs = "t,px,py,pz,qx,qy,qz,qw,vx,vy,vz,bax,bay,baz,bwx,bwy,bwz,cx,cy,cz,cdx,cdy,cdz,Lx,Ly,Lz,bpx,bpy,bpz\n";
    file_kfs << header_kfs;
    std::string header_cov = "t,Qpx,Qpy,Qpz,Qqx,Qqy,Qqz,Qvx,Qvy,Qvz,Qbax,Qbay,Qbaz,Qbwx,Qbwy,Qbwz,Qcx,Qcy,Qcz,Qcdx,Qcdy,Qcdz,QLx,QLy,QLz,Qbpx,Qbpy,Qbpz\n";
    file_cov << header_cov; 

    VectorComposite state_est;
    double o_p_ob_carr[3*N];
    double o_q_b_carr[4*N];
    double o_v_ob_carr[3*N];
    double o_p_oc_carr[3*N];
    double o_v_oc_carr[3*N];
    double o_Lc_carr[3*N];
    for (unsigned int i=0; i < N; i++) { 
        state_est = problem->getState(t_arr[i]);
        Vector3d i_omg_oi = i_omg_oi_v[i];
        Vector3d o_p_oi = state_est['P'];
        Vector4d o_qvec_i = state_est['O'];
        Vector3d o_v_oi = state_est['V'];

        Vector6d bias_acc_gyro = sen_imu->getStateBlockDynamic('I', t_arr[i])->getState();

        Matrix3d o_R_i = Quaterniond(o_qvec_i).toRotationMatrix();
        Matrix3d o_R_b = o_R_i * i_R_b;
        Vector4d o_qvec_b = Quaterniond(o_R_b).coeffs();
        // std::cout << "o_p_oi:                  " << o_p_oi.transpose() << std::endl;
        // std::cout << "i_p_ib:                  " << i_p_ib.transpose() << std::endl;
        // std::cout << "o_R_i * i_p_ib:          " << (o_R_i * i_p_ib).transpose() << std::endl;
        // std::cout << "o_p_oi + o_R_i * i_p_ib: " << (o_p_oi + o_R_i * i_p_ib).transpose() << std::endl;
        Vector3d o_p_ob = o_p_oi + o_R_i * i_p_ib;
        Vector3d o_v_ob = o_v_oi + o_R_b * (i_omg_oi - bias_acc_gyro.tail<3>()).cross(b_p_bi);

        Vector3d o_p_oc = state_est['C'];
        Vector3d o_v_oc = state_est['D'];
        Vector3d o_Lc = state_est['L'];
        
        mempcpy(o_p_ob_carr+3*i, o_p_ob.data(),   3*sizeof(double));
        mempcpy(o_q_b_carr +4*i, o_qvec_b.data(), 4*sizeof(double));
        mempcpy(o_v_ob_carr+3*i, o_v_ob.data(),   3*sizeof(double));
        mempcpy(o_p_oc_carr+3*i, o_p_oc.data(),   3*sizeof(double));
        mempcpy(o_v_oc_carr+3*i, o_v_oc.data(),   3*sizeof(double));
        mempcpy(o_Lc_carr  +3*i, o_Lc.data(),     3*sizeof(double));

        file_est << std::setprecision(std::numeric_limits<long double>::digits10 + 1)
                 << t_arr[i] << ","
                 << o_p_ob(0) << ","
                 << o_p_ob(1) << ","
                 << o_p_ob(2) << "," 
                 << o_qvec_b(0) << "," 
                 << o_qvec_b(1) << "," 
                 << o_qvec_b(2) << "," 
                 << o_qvec_b(3) << "," 
                 << o_v_ob(0) << "," 
                 << o_v_ob(1) << "," 
                 << o_v_ob(2) << "," 

                 << o_p_oc(0) << ","
                 << o_p_oc(1) << ","
                 << o_p_oc(2) << "," 
                 << o_v_oc(0) << "," 
                 << o_v_oc(1) << "," 
                 << o_v_oc(2) << "," 
                 << o_Lc(0) << "," 
                 << o_Lc(1) << "," 
                 << o_Lc(2)
                 << "\n"; 
    }


    VectorComposite kf_state;
    CaptureBasePtr cap_ikin;
    VectorComposite bp_bias_est;
    CaptureBasePtr cap_imu;
    VectorComposite bi_bias_est;
    for (auto elt: problem->getTrajectory()->getFrameMap())
    {
        auto kf = elt.second;
        kf_state = kf->getState();
        cap_ikin = kf->getCaptureOf(sen_ikin); 
        bp_bias_est = cap_ikin->getState();
        cap_imu = kf->getCaptureOf(sen_imu); 
        bi_bias_est = cap_imu->getState();

        file_kfs << std::setprecision(std::numeric_limits<long double>::digits10 + 1)
                    << kf->getTimeStamp().get() << ","
                    << kf_state['P'](0) << ","
                    << kf_state['P'](1) << ","
                    << kf_state['P'](2) << "," 
                    << kf_state['O'](0) << "," 
                    << kf_state['O'](1) << "," 
                    << kf_state['O'](2) << "," 
                    << kf_state['O'](3) << "," 
                    << kf_state['V'](0) << "," 
                    << kf_state['V'](1) << "," 
                    << kf_state['V'](2) << "," 
                    << bi_bias_est['I'](0) << ","
                    << bi_bias_est['I'](1) << ","
                    << bi_bias_est['I'](2) << ","
                    << bi_bias_est['I'](3) << ","
                    << bi_bias_est['I'](4) << ","
                    << bi_bias_est['I'](5) << ","
                    << kf_state['C'](0) << "," 
                    << kf_state['C'](1) << "," 
                    << kf_state['C'](2) << "," 
                    << kf_state['D'](0) << "," 
                    << kf_state['D'](1) << "," 
                    << kf_state['D'](2) << "," 
                    << kf_state['L'](0) << "," 
                    << kf_state['L'](1) << "," 
                    << kf_state['L'](2) << ","
                    << bp_bias_est['I'](0) << ","
                    << bp_bias_est['I'](1) << ","
                    << bp_bias_est['I'](2)
                    << "\n"; 

        // compute covariances of KF and capture stateblocks
        Eigen::MatrixXd Qp =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qo =  Eigen::Matrix3d::Identity();  // global or local?
        Eigen::MatrixXd Qv =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qbi =  Eigen::Matrix6d::Identity();
        Eigen::MatrixXd Qc =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qd =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd QL =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qbp = Eigen::Matrix3d::Identity();
        // solver->computeCovariances(kf->getStateBlockVec());     // !! does not work as intended...
        
        solver->computeCovariances(SolverManager::CovarianceBlocksToBeComputed::ALL);  // should not be computed every time !! see below
        problem->getCovarianceBlock(kf->getStateBlock('P'), Qp);
        problem->getCovarianceBlock(kf->getStateBlock('O'), Qo);
        problem->getCovarianceBlock(kf->getStateBlock('V'), Qv);
        problem->getCovarianceBlock(kf->getStateBlock('C'), Qc);
        problem->getCovarianceBlock(kf->getStateBlock('D'), Qd);
        problem->getCovarianceBlock(kf->getStateBlock('L'), QL);

        solver->computeCovariances(cap_ikin->getStateBlockVec());  // not computed by ALL and destroys other computed covs -> issue to raise
        problem->getCovarianceBlock(cap_ikin->getSensorIntrinsic(), Qbp);
        // std::cout << "Qbp\n" << Qbp << std::endl;
        solver->computeCovariances(cap_imu->getStateBlockVec());  // not computed by ALL and destroys other computed covs -> issue to raise
        problem->getCovarianceBlock(cap_imu->getSensorIntrinsic(), Qbi);

        file_cov << std::setprecision(std::numeric_limits<long double>::digits10 + 1)
                        << kf->getTimeStamp().get() << ","
                        << Qp(0,0) << ","
                        << Qp(1,1) << ","
                        << Qp(2,2) << ","
                        << Qo(0,0) << ","
                        << Qo(1,1) << ","
                        << Qo(2,2) << ","
                        << Qv(0,0) << ","
                        << Qv(1,1) << ","
                        << Qv(2,2) << ","
                        << Qbi(0,0) << ","
                        << Qbi(1,1) << ","
                        << Qbi(2,2) << ","
                        << Qbi(3,3) << ","
                        << Qbi(4,4) << ","
                        << Qbi(5,5) << ","
                        << Qc(0,0) << ","
                        << Qc(1,1) << ","
                        << Qc(2,2) << ","
                        << Qd(0,0) << ","
                        << Qd(1,1) << ","
                        << Qd(2,2) << ","
                        << QL(0,0) << ","
                        << QL(1,1) << ","
                        << QL(2,2) << ","
                        << Qbp(0,0) << ","
                        << Qbp(1,1) << ","
                        << Qbp(2,2)
                        << "\n"; 
    }

    for (unsigned int i=0; i < N; i++) { 
        file_fbk << std::setprecision(std::numeric_limits<long double>::digits10 + 1)
                 << t_arr[i] << ","
                 << p_ob_fbk_v[i](0) << ","
                 << p_ob_fbk_v[i](1) << ","
                 << p_ob_fbk_v[i](2) << "," 
                 << q_ob_fbk_v[i](0) << "," 
                 << q_ob_fbk_v[i](1) << "," 
                 << q_ob_fbk_v[i](2) << "," 
                 << q_ob_fbk_v[i](3) << "," 
                 << v_ob_fbk_v[i](0) << "," 
                 << v_ob_fbk_v[i](1) << "," 
                 << v_ob_fbk_v[i](2) << ","
                 << c_ob_fbk_v[i](0) << "," 
                 << c_ob_fbk_v[i](1) << "," 
                 << c_ob_fbk_v[i](2) << "," 
                 << cd_ob_fbk_v[i](0) << "," 
                 << cd_ob_fbk_v[i](1) << "," 
                 << cd_ob_fbk_v[i](2) << "," 
                 << Lc_ob_fbk_v[i](0) << "," 
                 << Lc_ob_fbk_v[i](1) << "," 
                 << Lc_ob_fbk_v[i](2)
                 << "\n";    
    }

    file_est.close();
    file_kfs.close();
    file_cov.close();


    // Save trajectories in npz file
    // save ground truth
    cnpy::npz_save(out_npz_file_path, "t", t_arr, {N}, "w");             // "w" overwrites any file with same name
    cnpy::npz_save(out_npz_file_path, "w_p_wm", w_p_wm_arr, {N,3}, "a"); // "a" appends to the file we created above
    cnpy::npz_save(out_npz_file_path, "w_q_m",  w_q_m_arr,  {N,4}, "a"); 
    cnpy::npz_save(out_npz_file_path, "w_v_wm", w_v_wm_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "m_v_wm", m_v_wm_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_i", o_q_i_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "qa", qa_arr, {N,12}, "a");

    // save estimates
    cnpy::npz_save(out_npz_file_path, "o_p_ob", o_p_ob_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_b",  o_q_b_carr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_ob", o_v_ob_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_p_oc", o_p_oc_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_oc", o_v_oc_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_Lc",   o_Lc_carr,   {N,3}, "a");

}
