//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file gtest_capture_leg_odom.cpp
 *
 *  Created on: May 12, 2020
 *      \author: mfourmy
 */


#include "bodydynamics/capture/capture_leg_odom.h"

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

#include <Eigen/Dense>

using namespace wolf;
using namespace Eigen;


TEST(CaptureLegOdom, point_feet_relative_kin)
{   
  TimeStamp ts(0.0);
  double dt = 1;
  // 2 point feet + wb
  MatrixXd data_kin(3,5);
  Vector3d bm_p_bml1; bm_p_bml1 << 0,0,0;
  Vector3d b_p_bl1; b_p_bl1 << -1,0,0;
  Vector3d bm_p_bml2; bm_p_bml2 << 0,0,0;
  Vector3d b_p_bl2; b_p_bl2 << -1,0,0;
  Vector3d i_omg_oi; i_omg_oi << 0,0,0;
  data_kin.col(0) = bm_p_bml1;
  data_kin.col(1) = b_p_bl1;
  data_kin.col(2) = bm_p_bml2;
  data_kin.col(3) = b_p_bl2;
  data_kin.col(4) = i_omg_oi;

  Matrix6d data_cov = Matrix6d::Identity();
  CaptureLegOdomPtr CLO = std::make_shared<CaptureLegOdom>(ts, nullptr, data_kin, data_cov, dt, CaptureLegOdomType::POINT_FEET_RELATIVE_KIN);

  Vector3d bm_p_bmb_exp; bm_p_bmb_exp << 1, 0, 0;
  ASSERT_MATRIX_APPROX(CLO->getData().head<3>(), bm_p_bmb_exp, 1e-6)
}

TEST(CaptureLegOdom, point_feet_differential_kin)
{   
  TimeStamp ts(0.0);
  double dt = 1;
  // 2 point feet + wb
  MatrixXd data_kin(10,3);
  Vector3d i_omg_oi; i_omg_oi << 0,1,0;

  // 1rst foot: only relative foot/base motion from J*dqa, no lever so no influence from angular vel 
  Vector3d b_p_bl1; b_p_bl1 << 0,0,0; 
  Vector4d b_qvec_l1; b_qvec_l1 << 0,0,0,1; 
  Vector3d l_v_bl1; l_v_bl1 << -1,0,0;

  // 2nd feet: rigid rotation pivoting around the foot point
  Vector3d b_p_bl2; b_p_bl2 << 0,0,-1; 
  Vector4d b_qvec_l2; b_qvec_l2 << 0,0,0,1; 
  Vector3d l_v_bl2; l_v_bl2 << 0,0,0; 
  data_kin.col(0) << b_p_bl1, b_qvec_l1, l_v_bl1;
  data_kin.col(1) << b_p_bl2, b_qvec_l2, l_v_bl2;
  data_kin.col(2) << i_omg_oi, 0,0,0,0,0,0,0;

  Matrix6d data_cov = Matrix6d::Identity();
  CaptureLegOdomPtr CLO = std::make_shared<CaptureLegOdom>(ts, nullptr, data_kin, data_cov, dt, CaptureLegOdomType::POINT_FEET_DIFFERENTIAL_KIN);

  Vector3d bm_p_bmb_exp; bm_p_bmb_exp << 1, 0, 0;
  ASSERT_MATRIX_APPROX(CLO->getData().head<3>(), bm_p_bmb_exp, 1e-6)
}

TEST(CaptureLegOdom, humanoid_feet)
{   
  TimeStamp ts(0.0);
  double dt = 1;
  // 1 humanoid feet + wb
  MatrixXd data_kin(7,2); 
  // only one foot taken into account
  Vector3d bm_p_bml1; bm_p_bml1 << 0,0,0;
  Vector4d bm_q_bml1; bm_q_bml1 << 0,0,0,1;
  Vector3d b_p_bl1; b_p_bl1 << -1,0,0;
  Vector4d b_q_bl1; b_q_bl1 << 0,0,0,1;
  data_kin.col(0) << bm_p_bml1, bm_q_bml1;
  data_kin.col(1) << b_p_bl1, b_q_bl1;

  Matrix6d data_cov = Matrix6d::Identity();
  CaptureLegOdomPtr CLO = std::make_shared<CaptureLegOdom>(ts, nullptr, data_kin, data_cov, dt, CaptureLegOdomType::HUMANOID_FEET_RELATIVE_KIN);

  Vector3d bm_p_bmb_exp; bm_p_bmb_exp << 1, 0, 0;
  ASSERT_MATRIX_APPROX(CLO->getData().head<3>(), bm_p_bmb_exp, 1e-6)

}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}

