//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#include "bodydynamics/factor/factor_force_torque_inertial_dynamics.h"
#include "bodydynamics/processor/processor_force_torque_inertial_dynamics.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"
#include "bodydynamics/internal/config.h"

#include <core/sensor/factory_sensor.h>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>
#include <core/utils/params_server.h>
#include <core/yaml/parser_yaml.h>
#include <core/state_block/state_block_derived.h>
#include <core/state_block/factory_state_block.h>

#include <Eigen/Dense>
#include <Eigen/Geometry>

using namespace wolf;
using namespace bodydynamics::fti;

WOLF_REGISTER_STATEBLOCK_WITH_KEY(L, StateVector3d);

class Test_ProcessorForceTorqueInertialDynamics_yaml : public testing::Test
{
  public:
    ProblemPtr                                    P;
    SensorForceTorqueInertialPtr                  S;
    ProcessorForceTorqueInertialDynamicsPtr p;

  protected:
    void SetUp() override
    {
        std::string  wolf_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
        ParserYaml   parser    = ParserYaml("problem_force_torque_inertial_dynamics_processor_test.yaml", wolf_root + "/test/yaml/");
        ParamsServer server    = ParamsServer(parser.getParams());
        P                      = Problem::autoSetup(server);

        S = std::static_pointer_cast<SensorForceTorqueInertial>(P->getHardware()->getSensorList().front());
        p = std::static_pointer_cast<ProcessorForceTorqueInertialDynamics>(S->getProcessorList().front());
    }
};

TEST_F(Test_ProcessorForceTorqueInertialDynamics_yaml, force_registraion)
{
    FactorySensor::registerCreator("SensorForceTorqueInertial", SensorForceTorqueInertial::create);
}

// Test to see if the processor works (yaml files). Here the dron is not moving
TEST_F(Test_ProcessorForceTorqueInertialDynamics_yaml, not_moving_test)
{
    VectorXd data             = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0)        = - gravity();
    data.segment<3>(6)        = - 1.952*gravity();
    MatrixXd         data_cov = MatrixXd::Identity(12, 12);

    CaptureMotionPtr C0_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1       = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());


    //We check that, effectively, the dron has not moved
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('P')->getState(), C_KF0->getFrame()->getStateBlock('P')->getState(), 1e-8);
    ASSERT_QUATERNION_APPROX(Quaterniond(C_KF1->getFrame()->getStateBlock('O')->getState().data()), Quaterniond(C_KF0->getFrame()->getStateBlock('O')->getState().data()), 1e-8);
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('V')->getState(), C_KF0->getFrame()->getStateBlock('V')->getState(), 1e-8);
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('L')->getState(), C_KF0->getFrame()->getStateBlock('L')->getState(), 1e-8);


    //Checking that the captures we have taken are the correct ones by looking both time stands 
    ASSERT_EQ(C_KF0->getTimeStamp(), C0_0->getTimeStamp());
    ASSERT_EQ(C_KF1->getTimeStamp(), C5_1->getTimeStamp());


    Vector3d p0 = C_KF0->getFrame()->getStateBlock('P')->getState();
    Quaterniond q0 = Quaterniond(C_KF0->getFrame()->getStateBlock('O')->getState().data());
    Vector3d v0 = C_KF0->getFrame()->getStateBlock('V')->getState();
    Vector3d L0 = C_KF0->getFrame()->getStateBlock('L')->getState();

    Vector3d p1 = C_KF1->getFrame()->getStateBlock('P')->getState();
    Quaterniond q1 = Quaterniond(C_KF1->getFrame()->getStateBlock('O')->getState().data());
    Vector3d v1 = C_KF1->getFrame()->getStateBlock('V')->getState();
    Vector3d L1 = C_KF1->getFrame()->getStateBlock('L')->getState();

    Vector3d dpi;
    Vector3d dvi;
    Vector3d dpd;
    Vector3d dvd;
    Vector3d dL;
    Quaterniond dq;
    double dt = 1.0;
    VectorXd betw_states(19);

    betweenStates(p0,v0,L0,q0,p1,v1,L1,q1,dt,dpi,dvi,dpd,dvd,dL,dq);

    betw_states << dpi, dvi, dpd, dvd, dL, dq.coeffs(); 
    
    VectorXd delta_preintegrated = C_KF1->getFeatureList().back()->getMeasurement();

    //Cheking that the preintegrated delta calculated by the processor is the same as the one that is correct for sure
    ASSERT_MATRIX_APPROX(betw_states, delta_preintegrated, 1e-8);

}

//Test to see if the processor works (yaml files). Here the dron is moving just 1m in x direction
TEST_F(Test_ProcessorForceTorqueInertialDynamics_yaml, 1m_x_moving_test__com_zero)
{
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data(0)            = 2;
    data.segment<3>(6) = - 1.952*gravity();
    data(6)            = 1.952*2;
    MatrixXd         data_cov = MatrixXd::Identity(12, 12);

    // Set CoM to zero so that an acceleration on X does not produce a torque, thereby producing non-null 'L' at KF1
    S->getStateBlock('C')->setState(Vector3d::Zero());

    CaptureMotionPtr C0_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.3, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0       = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);
    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    P->print(4,1,1,1);

    C_KF1->getBuffer().print(1,1,1,0,0);

    //We check that, effectively, the drone has moved 1m in the x direction, the x component of the
    //velocity is now 2m/s and nothing else has changed from the original state
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('P')->getState(), Vector3d(1,0,0), 1e-8);
    ASSERT_QUATERNION_APPROX(Quaterniond(C_KF1->getFrame()->getStateBlock('O')->getState().data()), Quaterniond(C_KF0->getFrame()->getStateBlock('O')->getState().data()), 1e-8);
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('V')->getState(), Vector3d(2,0,0), 1e-8);
    ASSERT_MATRIX_APPROX(C_KF1->getFrame()->getStateBlock('L')->getState(), C_KF0->getFrame()->getStateBlock('L')->getState(), 1e-8);



    Vector3d p0 = C_KF0->getFrame()->getStateBlock('P')->getState();
    Quaterniond q0 = Quaterniond(C_KF0->getFrame()->getStateBlock('O')->getState().data());
    Vector3d v0 = C_KF0->getFrame()->getStateBlock('V')->getState();
    Vector3d L0 = C_KF0->getFrame()->getStateBlock('L')->getState();

    Vector3d p1 = C_KF1->getFrame()->getStateBlock('P')->getState();
    Quaterniond q1 = Quaterniond(C_KF1->getFrame()->getStateBlock('O')->getState().data());
    Vector3d v1 = C_KF1->getFrame()->getStateBlock('V')->getState();
    Vector3d L1 = C_KF1->getFrame()->getStateBlock('L')->getState();

    Vector3d dpi;
    Vector3d dvi;
    Vector3d dpd;
    Vector3d dvd;
    Vector3d dL;
    Quaterniond dq;
    double dt = 1.0;
    VectorXd betw_states(19);

    betweenStates(p0,v0,L0,q0,p1,v1,L1,q1,dt,dpi,dvi,dpd,dvd,dL,dq);

    betw_states << dpi, dvi, dpd, dvd, dL, dq.coeffs(); 
    
    VectorXd delta_preintegrated = C_KF1->getFeatureList().back()->getMeasurement();

    //Cheking that the preintegrated delta calculated by the processor is the same as the one that is correct for sure
    ASSERT_MATRIX_APPROX(betw_states, delta_preintegrated, 1e-8);

}

//Test to see if the voteForKeyFrame works (distance traveled)
TEST_F(Test_ProcessorForceTorqueInertialDynamics_yaml, VoteForKeyFrame_dist)
{
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data(0)            = 2;
    data.segment<3>(6) = - 1.952*gravity();
    data(6)            = 1.952*2;
    MatrixXd         data_cov = MatrixXd::Identity(12, 12);

    p->setMaxTimeSpan(999);
    p->setDistTraveled(0.995);

    CaptureMotionPtr C0_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.3, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0       = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);
    C0_0->process();
    C1_0->process();
    C2_0->process();
    C3_0->process();

    P->print(4,1,1,1);

}

//Test to see if the voteForKeyFrame works (buffer)
TEST_F(Test_ProcessorForceTorqueInertialDynamics_yaml, VoteForKeyFrame_buffer)
{
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data(0)            = 2;
    data.segment<3>(6) = - 1.952*gravity();
    data(6)            = 1.952*2;
    MatrixXd         data_cov = MatrixXd::Identity(12, 12);

    p->setMaxTimeSpan(999);
    p->setMaxBuffLength(3);

    CaptureMotionPtr C0_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.3, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0       = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0       = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);
    C0_0->process();
    C1_0->process();
    C2_0->process();
    C3_0->process();

    P->print(4,1,1,1);

}


int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    // ::testing::GTEST_FLAG(filter) = "Test_ProcessorForceTorqueInertialDynamics_yaml.not_moving_test";
    return RUN_ALL_TESTS();
}